/*
 *   Copyright (C) 2002-2006
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "tac_plus.h"
#include "tacacs.h"
#include <stdarg.h>
#include <unistd.h>

db::db(string *url) {
	dbh = new DBH;
	setcfg(url);

	connected = 0;
}

db::db( const db& x)
{
	dbh = new DBH;
	type = x.type;
	user = x.user;
	passwd = x.passwd;
	host = x.host;
	database = x.database;

	connected = 0;
}

db::~db() {
	delete dbh;
}

int db::connect()
{
	if (connected)
		return (0);

#if defined(DB_MYSQL)
	if (!mysql_init(dbh)) {
		if (taclog->debug)
			taclog->report("db::connect - Cannot initialize MYSQL db\n");
		return (1);
	}

	if (!mysql_real_connect(dbh, host.c_str(), user.c_str(), passwd.c_str(), database.c_str(), 0, NULL, 0)) {
		if (taclog->debug)
			taclog->report("db::connect - Cannot connect to MYSQL database\n");
		return (1);
	}
#endif

#if defined(DB_PGSQL)
	dbh->conn = PQsetdbLogin(host.c_str(), NULL, NULL, NULL, database.c_str(), user.c_str(), passwd.c_str());
	if (PQstatus(dbh->conn) == CONNECTION_BAD) {
		if (taclog->debug)
			taclog->erport("db::connect - Cannot connect to PG database");
		return (1);
	}
#endif

	connected++;
	return (0);
}

void db::close()
{
#if defined(DB_MYSQL)
	mysql_close(dbh);
#endif

#if defined(DB_PGSQL)
	PQfinish(dbh->conn);
#endif
	connected=0;
}

DB_RES *db::store_result()
{
#if defined(DB_MYSQL)
        return mysql_store_result(dbh);
#endif

#if defined(DB_PGSQL)
        return dbh->res;
#endif
}

int db::num_rows(DB_RES *result)
{
#if defined(DB_MYSQL)
        return mysql_num_rows(result);
#endif

#if defined(DB_PGSQL)
        return PQntuples(result);
#endif
}

DB_ROW db::fetch_row(DB_RES *result)
{
#if defined(DB_MYSQL)
        return mysql_fetch_row(result);
#endif
}

void db::free_result(DB_RES *result)
{
#if defined(DB_MYSQL)
        mysql_free_result(result);
#endif

#if defined(DB_PGSQL)
        PQclear(result);
#endif
}

int db::affected_rows()
{
#if defined(DB_MYSQL)
        return mysql_affected_rows(dbh);
#endif
}

int db::query(string *querystmt)
{
	int ret = 0;

#if defined(DB_MYSQL)
        if (mysql_real_query(dbh, querystmt->c_str(), querystmt->size())) {
                if (taclog->debug)
			taclog->report("db::query - Cannot query using MySQL. Error - %s", mysql_error(dbh));
                ret = 1;
        }
#endif

#if defined(DB_PGSQL)
        dbh->res = PQexec(dbh->conn, querystmt->c_str());
        if (!dbh->res || PQresultStatus(dbh->res) != PGRES_COMMAND_OK )
        {
                if (taclog->debug)
			taclog->report("db::query: Cannot query using PGSQL.  Error - %s", PQerrorMessage(dbh->conn));
                ret = 1;
        }
#endif

        return(ret);
}

#ifdef __STDC__
int db::query(char *querystmt,...)
#else
int db::query(querystmt, va_alist)
char *querystmt;
va_dcl
#endif
{
        char *p, *sval, *sqlcmd, *bufp;
        char digits[16];
        int m, len = 0, ret = 0;
        va_list ap;

        bufp = sqlcmd = new char[SQLCMDL]; //(char*)malloc(SQLCMDL);
#ifdef __STDC__
	va_start(ap, querystmt);
#else
	va_start(ap);
#endif

        for (p = querystmt; *p; p++) {
                if (*p != '%') {
                        *bufp++ = *p;
                        len++;
                        continue;
                }
                switch (*++p) {
                case 's':
                        sval = va_arg(ap, char *);
                        m = strlen(sval);
                        break;

                case 'u':
                        sprintf(digits, "%u", va_arg(ap, uint));
                        m = strlen(digits);
                        sval = digits;
                        break;

                case 'x':
                        sprintf(digits, "%x", va_arg(ap, uint));
                        m = strlen(digits);
                        sval = digits;
                        break;

                case 'd':
                        sprintf(digits, "%d", va_arg(ap, int));
                        m = strlen(digits);
                        sval = digits;
                        break;
                }

                memcpy(bufp, sval, m);
                bufp += m;
                len += m;
                continue;
        }

        sqlcmd[len] = '\0';
        va_end(ap);

#if defined(DB_MYSQL)
        if (mysql_real_query(dbh, sqlcmd, len)) {
                if (taclog->debug)
			taclog->report("db::query - Cannot query (%s) using MySQL. Error - %s\n", sqlcmd, mysql_error(dbh));
                ret = 1;
        }
#endif

#if defined(DB_PGSQL)
        dbh->res = PQexec(dbh->conn, sqlcmd);
        if (!dbh->res || PQresultStatus(dbh->res) != PGRES_COMMAND_OK )
        {
                if (taclog->debug)
			taclog->report(LOG_DEBUG, "db::query: Cannot query using PGSQL.  Error - %s", PQerrorMessage(dbh->conn));
                ret = 1;
        }
#endif

        delete sqlcmd;
        return(ret);
}

int db::setcfg( string *url)
{
	size_t current=0, last=0;

	if (url->empty()) {
		if (taclog->debug & DEBUG_PARSE_FLAG)
			taclog->report("db::setcfg - Cannot use an empty url string\n");
		return (1);
	}

	if (taclog->debug & DEBUG_PARSE_FLAG) {
		taclog->report("db::setcfg - parsing %s\n",url->c_str());
	}

//	dburl->assign(url->c_str());
	try {
		last=url->find("://");
		type = url->substr(current, last);
		current = last+3;
	}  catch (exception e) {
		if (taclog->debug & DEBUG_PARSE_FLAG)
			taclog->report("db::setcfg - Cannot parse databse user\n");
		return (1);
	}

	try {
		last=url->find(":",current);
		user = url->substr(current, last-current);
		current = last+1;
	} catch (exception e) {
		if (taclog->debug & DEBUG_PARSE_FLAG)
			taclog->report("db::setcfg - Cannot find password\n");
		return (1);
	}

	try {
		last=url->find("@",current);
		passwd = url->substr(current, last-current);
		current = last+1;
	} catch (exception e) {
		if (taclog->debug & DEBUG_PARSE_FLAG)
			taclog->report("db::setcfg - Cannot find hostname\n");
		return (1);
	}

	try {
		last=url->find("/",current);
		host = url->substr(current, last-current);
	} catch (exception e) {
		if (taclog->debug & DEBUG_PARSE_FLAG)
			taclog->report("db::setcfg - Cannot find database\n");
		return (1);
	}

	database = url->substr(last+1);

	if (taclog->debug & DEBUG_PARSE_FLAG) {
		taclog->report("db::setcfg - dbtype=%s, dbuser=%s, dbpasswd=%s, dbhost=%s, db=%s\n", type.c_str(), user.c_str(), passwd.c_str(), host.c_str(), database.c_str());
	}

	return (0);
}

/*
int db::setcfg( char *url)
{
	char *pUrl;
	int len;

	if (!url) {
		taclog->report("db::setcfg - Cannot use a NULL url\n");
		return (1);
	}

	if (taclog->debug & DEBUG_PARSE_FLAG) {
		taclog->report("db::setcfg - parsing %s\n", url);
	}

	dburl = url;

	len = (strlen(url) + 1) * sizeof(char);
	pUrl = new char[len];
	bzero(pUrl, len);
	strcpy(pUrl, url);

	type = pUrl;

	if (NULL == (pUrl = (char*)strstr( pUrl, "://")) ) {
		taclog->report("db::setcfg - Cannot parse databse user\n");
		delete pUrl;
		return (1);
	}
	*pUrl = '\0';
	pUrl += 3;

	user = pUrl;

	if (NULL == (pUrl = (char*)strchr( pUrl, ':')) ) {
		taclog->report("db::setcfg - Cannot find password\n");
		delete pUrl;
		return (1);
	}

	*(pUrl++) = '\0';
	passwd = pUrl;

	if ( NULL == (pUrl = (char*)strchr( pUrl, '@' )) ) {
		taclog->report("db::setcfg - Cannot find hostname\n");
		delete pUrl;
		return (1);
	}

	*(pUrl++) = '\0';
	host = pUrl;

	if ( NULL == (pUrl = (char*)strchr( pUrl, '/')) ) {
		taclog->report("db::setcfg - Cannot find database\n");
		delete pUrl;
		return (1);
	}

	*(pUrl++) = '\0';
	database = pUrl;

	if (taclog->debug & DEBUG_PARSE_FLAG) {
		taclog->report("db::setcfg - dbtype=%s, dbuser=%s, dbpasswd=%s, dbhost=%s, db=%s\n", type, user, passwd, host, database);
	}

	return (0);
}
*/

int db::get_secret(string user, int type, string *secret, time_t t)
{
	int ret = 1;
	string atype;

	DB_RES *res;
	DB_ROW row;

	if (user.empty())
		return (2);

	if (taclog->debug)
		taclog->report("db::get_secret - getting secret for user(%s)", user.c_str());

	switch (type) {
	case TAC_PLUS_AUTHEN_TYPE_CHAP:
		atype = "chap";
		break;
	case TAC_PLUS_AUTHEN_TYPE_MSCHAP:
		atype = "mschap";
		break;
	case TAC_PLUS_AUTHEN_TYPE_ARAP:
		atype = "arap";
		break;
	case TAC_PLUS_AUTHEN_TYPE_PAP:
		atype = "pap";
		break;
	default:
		if (taclog->debug)
			taclog->report("db::get_secret - Authentication type is unknown\n");
		return (2);
	}

	if (connect())
		return (ret);

	if (query(USERSECRETSQL, atype.c_str(), user.c_str())) {
		res = store_result();
		if (num_rows(res) >= 1) {
			row = fetch_row(res);
			secret->assign(row[0]);
			t = atol(row[1]);
			ret = 0;
		}
		free_result(res);
	}

	close();

	if (!secret->empty())
		if (taclog->debug)
			taclog->report("tauthen::get_secret - Secret found (%s)\n", secret->c_str());

	return (ret);
}

int db::get_host(char *field, uint32_t nas, string *data)
{
	int ret = 1;

	DB_RES *res, *res1;
	DB_ROW row, row1;

	if (!field) {
		return(2);
	}

	if (taclog->debug)
		taclog->report("db::get_host: getting %s from nas(%u)\n", field, nas);

	if (connect())
		return (ret);

	if (!query(HOSTAUTHSQL, field, nas)) {
		res = store_result();
		if (num_rows(res) >= 1) {
			while (ret && (row = fetch_row(res))) {
				if (string(row[0]).empty()) {
					if (!query(HOSTGROUPSQL, field, row[1])) {
						res1 = store_result();
						if (num_rows(res1) > 0) {
							while (ret && (row1 = fetch_row(res1))) {
								if (string(row1[0]).empty()) {
									if (taclog->debug) taclog->report("db::get_host: hostgroup is empty\n");
								} else {
									if (taclog->debug) taclog->report("db::get_host: got (%s) from hostgroup (%s)\n", row1[0], row[1]);
									data->assign(row1[0]);
									ret = 0;
								}
							}
							free_result(res1);
						}
					} else {
						if (taclog->debug)
							taclog->report("db::get_host: nasgroup(%s) not found\n", row[1]);
					}
				} else {
					if (taclog->debug) taclog->report("db::get_host: got (%s)\n", row[0]);
					data->assign(row[0]);
					ret = 0;
				}
			}
		} else {
			if (taclog->debug)
				taclog->report("db::get_host: Cannot find host\n");
		}
		free_result(res);
	}

	close();

	return (ret);
}

int db::get_user(struct identity *id, struct author_user *user)
{
	int ret = 1;

	DB_RES *res;
	DB_ROW row;

	if (connect())
		return (ret);

	if (!query("SELECT b_author, a_author, svc_dflt, cmd_dflt, gid FROM user WHERE uid='%s' and user=1", id->username->c_str())) {
		res = store_result();
		if (num_rows(res)>0) {
			row = fetch_row(res);
			user->pre_author = row[0]?row[0]:"";
			user->post_author = row[1]?row[0]:"";
			user->svc_dflt = atoi(row[2]?row[2]:"0");
			user->cmd_dflt = atoi(row[3]?row[3]:"0");
			user->member = row[4]?row[4]:"";
			free_result(res);

			ret = 0;

			if (!user->member.empty() && !(user->svc_dflt & user->cmd_dflt)) {
				if (!query("SELECT svc_dflt, cmd_dflt FROM user WHERE uid='%s' AND user=2", user->member.c_str())) {
					res = store_result();
					if (num_rows(res)>0) {
						user->svc_dflt=user->svc_dflt?user->svc_dflt:atoi(row[0]);
						user->cmd_dflt=user->cmd_dflt?user->cmd_dflt:atoi(row[1]);
						free_result(res);
					}
				}
			}

			if (!get_profile(id, user->member, 2, &user->profile, &user->vid)) {
				if (user->profile && !query("SELECT uid FROM user WHERE id=%u AND user=3", user->profile)) {
					res = store_result();
					if (num_rows(res)) {
						row = fetch_row(res);
						user->username = string(row[0]);
						user->member.erase();
						free_result(res);
					}
				}
			}
		}
	}

	return (ret);
}

int db::acct( string *querycmd )
{
	int ret = 0;

	if (taclog->debug) {
		taclog->report("db:acct - logging user access\n");
		taclog->report(querycmd);
	}

        if (!connect()) {
                if (!query(querycmd)) {
                        if (affected_rows()>0) {
                                ret = 1;
                                if (taclog->debug & DEBUG_ACCT_FLAG)
                                        taclog->report("db::access - Inserting record is successfull\n");
                        }
                }
        }

        close();
        return (ret);
}

int
db::verify(char *user_passwd, struct identity *id, int auth_type)
{
	int cont=1, i;
	string member, password, field_passwd;
	DB_RES *res;
	DB_ROW row;

	if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
		taclog->report("db::verify - Verify user (%s) to %s database\n", id->username->c_str(), type.c_str());

	if (auth_type == S_enable) {
		field_passwd = "enable";
	} else {
		field_passwd = "password";
	}

	if (connect()) {
		if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			taclog->report("db::verify - Cannot connected to the database\n");
		return (TAC_PLUS_AUTHEN_STATUS_ERROR);
	}

	if (query(AUTHSQL, field_passwd.c_str(), id->username->c_str())) {
		if (taclog->debug)
			taclog->report("db::verify - database problems\n");
		close();
		return(TAC_PLUS_AUTHEN_STATUS_ERROR);
	}

	res = store_result();
	if (num_rows(res) == 0) {
		if (taclog->debug)
			taclog->report("db::verify - user(%s) not found\n", id->username->c_str());
		close();
		return(TAC_PLUS_AUTHEN_STATUS_FAIL);
	}
	row = fetch_row(res);
	if (atoi(row[2])) {
		if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			taclog->report("db::verify - user(%s) is disabled\n", id->username->c_str());
		free_result(res);
		close();
		return (TAC_PLUS_AUTHEN_STATUS_FAIL);
	}

	password = row[0];
	member = row[1];
	i = atoi(row[3]);
	free_result(res);

	if ((i>1)) { //external authentication
		switch (i) {
			case 5: //SecurID authentication
				if (SD_Authenticate(*id->username, user_passwd)>0) {
					close();
					return (TAC_PLUS_AUTHEN_STATUS_FAIL);	
				}
				break;
			default:
				if (taclog->debug)
					taclog->report("db::verify - Feature not implemented\n");
				close();
				return (TAC_PLUS_AUTHEN_STATUS_FAIL);
		}
	} else {
	    if (password.empty()) {
		if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			taclog->report("db::verify - user password is empty\n");

		if (member.empty()) {
			if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
				taclog->report("db::verify - User(%s) is not a member of any group\n", id->username->c_str());
			close();
			return(TAC_PLUS_AUTHEN_STATUS_FAIL);
   		}

		if (query(AUTHSQL, field_passwd.c_str(), member.c_str())) {
			close();
			return(TAC_PLUS_AUTHEN_STATUS_FAIL);
		}

		res = store_result();
		if (num_rows(res) == 0) {
			if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
				taclog->report("db::verify - Member(%s) not found\n", member.c_str());
			close();
			return(TAC_PLUS_AUTHEN_STATUS_FAIL);
   		}
		row = fetch_row(res);
		password.assign(row[0]);
		free_result(res);

		if (password.empty()) {
			if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
				taclog->report("db::verify - Member password is blank.  Not allowed\n");
			close();
			return(TAC_PLUS_AUTHEN_STATUS_FAIL);
		}
	    }
	}

   // Verify if user is allowed to access the NAS
	if (get_profile(id, member, auth_type, (uint32_t*)&i, &i)) {
		if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			taclog->report("db::verify - User(%s) or member(%s) is not authorized to use NAS(%u)\n", id->username->c_str(), member.c_str(), id->NAS_addr);
		close();
		return (TAC_PLUS_AUTHEN_STATUS_FAIL);
	}

   // Verify if user is allowed to access the NAS from current IP
	if (!id->NAC_address->empty()) {
		if (query(USERACCESSSQL, id->username->c_str(), member.c_str(), id->NAC_address->c_str())) {
			if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
				taclog->report("db::verify - database problems.\n");
			close();
			return(TAC_PLUS_AUTHEN_STATUS_FAIL);
		}
		res = store_result();
		if (num_rows(res)) {
			cont = 1;
			while (cont && (row = fetch_row(res))) {
				if (strcmp(row[0],"58")==0) {
					if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
						taclog->report("db::verify - User is not authorized to access nas(%s) from this address(%s)\n", id->NAS_name->c_str(), id->NAC_address->c_str());
					free_result(res);
					close();
					return(TAC_PLUS_AUTHEN_STATUS_FAIL);
				} else if (strcmp(row[0],"57")==0)
					cont=0;
			}

			if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
				taclog->report("db::verify - User is authorized to access nas(%s) from this address(%s)\n", id->NAS_name->c_str(), id->NAC_address->c_str());
		}
	} else {
		if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			taclog->report("db::verify - User is login in through console\n");
	}
	// Verify password
	int status = TAC_PLUS_AUTHEN_STATUS_PASS;
	string *ep = new string(crypt(user_passwd, password.c_str()));
	if (!ep->compare(password)) {
		if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			taclog->report("db::verify - Password match\n");
	} else {
		if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
			taclog->report("db::verify - Password not match\n");
		status = TAC_PLUS_AUTHEN_STATUS_FAIL;
	}

	delete ep;
	close();
	return (status); 
}

int db::author( tauthor *data )
{

	int ret = 1, svc;
	DB_RES *res;
	DB_ROW row;
	string cmd, protocol, svcname;
	struct author_user user_rec;

	if (taclog->debug & DEBUG_AUTHOR_FLAG)
		taclog->report("db::author - authorizing user (%s)\n", data->id->username->c_str());
	user_rec.username.assign(data->id->username->c_str());
	data->status = AUTHOR_STATUS_ERROR;

	if (connect()) {
		if (taclog->debug & DEBUG_AUTHOR_FLAG)
			taclog->report("db::author - cannot access database\n");
		return (ret);
	}

	if (get_user(data->id, &user_rec)) {
		if (taclog->debug & DEBUG_AUTHOR_FLAG)
			taclog->report("db::author - failed to get user(%s) for authorization\n", data->id->username);
		close();
		return (ret);
	}

	if ( data->pre_authorization(user_rec.pre_author)) {
		close();
		return (0);
	}

	if (!(svc=data->get_nas_svc(&cmd, &protocol, &svcname))) {
		data->status = AUTHOR_STATUS_ERROR;
		data->admin_msg = new string("No identifiable service/protocol in authorization request");
		if (taclog->debug & DEBUG_AUTHOR_FLAG)
			taclog->report("db::author - %s for %s\n", data->admin_msg->c_str(), user_rec.username.c_str());
		close();
		return (0);
	}

	switch (svc) {
	default:
		if (taclog->debug & DEBUG_AUTHOR_FLAG)
			taclog->report("db::author - bad service type (%d) from %s\n", svc, data->sock->getForeignAddress().c_str());
		break;

	case N_svc_cmd:
		ret = authorize_cmd(&user_rec, cmd, data);
		break;

	case N_svc_exec:
		if (authorize_exec(&user_rec, data)) {
			close();
			return (0);
		}
						
	case N_svc_arap:
	case N_svc_ppp:
	case N_svc_slip:
	case N_svc:
		ret = authorize_svc(&user_rec, svc, protocol, svcname, data);
		break;
	}

	data->post_authorization(user_rec.post_author);

	if (taclog->debug & DEBUG_AUTHOR_FLAG)
		taclog->report("db::author - Finish authorization\n");

	close();
	return (ret);
}

int db::get_svc_node(string user, int vendor, int svc, string value, string value1)
{
	char *_vid;
	int ret = 0;

	if (taclog->debug & DEBUG_AUTHOR_FLAG) {
		taclog->report("db::get_svc_node - user (%s) service(%d) vendor(%d)\n", user.c_str(), svc, vendor);
	}

	// Check to see if the device was assigned a vendor
	_vid = new char[30];
	if (vendor) {
		sprintf(_vid,"AND (vid=0 or vid=%d)", vendor);
	} else {
		sprintf(_vid,"");
	}

	switch (svc) {
		case N_svc:
			if (query("SELECT type, value FROM node WHERE uid = '%s' AND service = %d AND value1 ='%s' %s", user.c_str(), svc, value1.c_str(), _vid))
				ret = 1;
			break;

		case N_svc_ppp:
			if (query("SELECT type, value FROM node WHERE id = '%s' AND service = %d AND value1 = '%s' %s", user.c_str(), svc, value1.c_str(), _vid) )
				ret = 1;
			break;
			
	        case N_svc_cmd:
		       if (query("SELECT type, value1 FROM node WHERE uid = '%s' AND service = %d AND value = '%s' AND '%s' REGEXP value1 %s ORDER BY seq", user.c_str(), svc, value.c_str(), value1.c_str(), _vid)) 
			       ret = 1;
		       break;
		case N_svc_exec:
		       if (query("SELECT type, CONCAT(value1,'=',value) AS value FROM node WHERE uid = '%s' AND service = %d %s", user.c_str(), svc, _vid))
			       ret = 1;
		       break;
		default:
		       if (query("SELECT type, value FROM node WHERE uid = '%s' AND service = %d %s", user.c_str(), svc, _vid) )
		                ret = 1;
	}

	delete _vid;
	return (ret);
}

int db::authorize_svc(struct author_user *user, int svc, string protocol, string svcname, tauthor *data)
{
	DB_RES *res;
	DB_ROW row;
	map<string,string> cfg_args;

	if (taclog->debug & DEBUG_AUTHOR_FLAG) {
		taclog->report("db::authorize_svc - service authorization for %s\n", user->username.c_str());
	}

	if (get_svc_node(user->username, user->vid, svc, protocol, svcname)) {
		data->status = AUTHOR_STATUS_ERROR;
	//	data->num_out_args = 0;
		return (1);
	}

	res = store_result();
	if (num_rows(res) <= 0) {
		free_result(res);

		if (user->member.empty()) goto NEXT;
		if (taclog->debug & DEBUG_AUTHOR_FLAG)
			taclog->report("db::authorise_svc - service authorization for %s\n", user->member.c_str());

		if (get_svc_node(user->member, user->vid, svc, protocol, svcname)) {
			data->status = AUTHOR_STATUS_ERROR;
//			data->num_out_args = 0;
			return (1);
		}

		res = store_result();
		if (num_rows(res) >= 1)
			goto NEXT2;
		free_result(res);
NEXT: 
		if (user->svc_dflt ==  1) {
			if (taclog->debug & DEBUG_AUTHOR_FLAG)
				taclog->report("db::authorize_svc - svc=%d protocol=%s svcname=%s not found, permitted by default\n",
					svc, protocol.c_str(),
					svcname.c_str()); 

			data->status = AUTHOR_STATUS_PASS_ADD;
//			data->num_out_args = 0;
//			data->output_args = NULL;
			return (0);
		}

		if (taclog->debug & DEBUG_AUTHOR_FLAG)
				taclog->report("db::authorize_svc - svc=%d protocol=%s not found, denied by default\n",
					svc, protocol.c_str()); 

		data->status = AUTHOR_STATUS_FAIL;
//		data->output_args = NULL;
//		data->num_out_args = 0;
		return (0);
	}
NEXT2:
	while (row = fetch_row(res)) {
		string *arg = new string(row[1]);
		int index = arg->find("=");
		
		if ((index<=0) && (atoi(row[0])==N_optarg))
			index = arg->find("*");
		cfg_args[arg->substr(0,index)]=arg->substr(index);
		taclog->report("db::authorize_svc - [%s]%s\n", arg->substr(0,index).c_str(), arg->substr(index).c_str());
		delete arg;
	}
		
	return (data->author_svc(cfg_args, 1));
}

int db::authorize_cmd(struct author_user *userid, string cmd, tauthor *data)
{
	DB_RES *res;
	DB_ROW row;
	string args;
	int cont = 1;

	args = string(data->input_args["cmd-arg"].substr(1));

	if (taclog->debug & DEBUG_AUTHOR_FLAG)
		taclog->report("db::authorize_cmd: command (%s %s) authorization for %s member of %s\n", cmd.c_str(), args.c_str(), userid->username.c_str(), userid->member.c_str());


	if (cmd.empty()) {
		data->status = AUTHOR_STATUS_ERROR;
		data->admin_msg = new string("No command found");
		taclog->report("db::authorize_cmd - %s - %s %s\n", data->sock->getForeignAddress().c_str(), cmd.c_str(), data->admin_msg->c_str());
		return (0);
	}

	if (get_svc_node(userid->username, userid->vid, N_svc_cmd, cmd, args)) {
		data->status = AUTHOR_STATUS_ERROR;
		return (1);
	}
	res = store_result();
	if (num_rows(res) <= 0) {
		if (!userid->member.empty()) {
			if (taclog->debug & DEBUG_AUTHOR_FLAG)
				taclog->report("db::authorize_cmd - command (%s %s) authorization for %s\n", cmd.c_str(), args.c_str(), userid->member.c_str());
			if (get_svc_node(userid->member, userid->vid, N_svc_cmd, cmd.c_str(), args.c_str())) {
				data->status = AUTHOR_STATUS_ERROR;
				return (1);
			}
			res = store_result();
			if (num_rows(res) >= 1)
				goto NEXT4;
		}

                if (userid->cmd_dflt ==  1) {
                        if (taclog->debug & DEBUG_AUTHOR_FLAG)
                                taclog->report("db::authorize_cmd - command(%s) not found, permitted by default\n", cmd.c_str());
                        data->status = AUTHOR_STATUS_PASS_ADD;
                        return (0);
                }

                if (taclog->debug & DEBUG_AUTHOR_FLAG)
			taclog->report("db::authorize_cmd - command(%s) not found, denied by default\n", cmd.c_str());

                data->status = AUTHOR_STATUS_FAIL;
                return (0);
	}

	// command exists.  check if command is DENIED or PERMITTED
NEXT4:
	row = fetch_row(res);
	switch (atoi(row[0])) {
	case N_permit:
		if (taclog->debug & DEBUG_AUTHOR_FLAG) 
			taclog->report("db::authorize_cmd - command(%s %s) permitted\n", cmd.c_str(), args.c_str());
		data->status = AUTHOR_STATUS_PASS_ADD;
		break;
	case N_deny:
		if (taclog->debug & DEBUG_AUTHOR_FLAG)
			taclog->report("db::authorize_cmd - %s %s denied\n", cmd.c_str(), args.c_str());
		data->status = AUTHOR_STATUS_FAIL;
		break;
	default:
		data->status = AUTHOR_STATUS_ERROR;
		data->admin_msg = new string("Server error illegal configuration");
		if (taclog->debug)
			taclog->report("db::authorize_cmd - %s: %s %s %s\n", data->sock->getForeignAddress().c_str(), cmd.c_str(), args.c_str(), data->admin_msg);
		break;
	}

	free_result(res);
	return (0);
}

int
db::authorize_exec(struct author_user *userid, tauthor *data)
{
	DB_RES *res;

	if (taclog->debug & DEBUG_AUTHOR_FLAG)
		taclog->report("db::authorize_exec - exec authorization for %s member of %s\n", userid->username.c_str(), userid->member.c_str());

	if (query("SELECT service FROM node WHERE uid IN ('%s','%s') and service IN (%d,%d)" , userid->username.c_str(), userid->member.c_str(), N_svc_exec, N_svc_cmd)) {
		data->status = AUTHOR_STATUS_ERROR;
		return (1);
	}
	res = store_result();
	if (num_rows(res) > 0) {
		if (taclog->debug & DEBUG_AUTHOR_FLAG)
			taclog->report("db::authorize_exec - exec is explicitly permitted\n");
		free_result(res);
		return (0);
	}
	free_result(res);
	if (userid->svc_dflt == 1) {
		if (taclog->debug & DEBUG_AUTHOR_FLAG)
			taclog->report("db::author - exec permitted by default\n");
		data->status = AUTHOR_STATUS_PASS_ADD;
		return(1);
	}

	if (taclog->debug & DEBUG_AUTHOR_FLAG)
		taclog->report("db::authorize_exec - exec denied by default\n");

	data->status = AUTHOR_STATUS_FAIL;
	return (1);
}


int
db::get_profile(struct identity *id, string member, int auth_type, uint32_t *profile, int *vendor)
{
	int num_row, i, acl[15], cont=1;
	string hostgroup;
	DB_RES *res, *res1;
	DB_ROW row, row1;

	if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG|DEBUG_AUTHOR_FLAG))
		taclog->report("db::get_profile - authorizing user (%s) member of (%s)\n", id->username->c_str(), member.c_str());

	if (connect())
		return (cont);

   	if (!query(HOSTACLSQL,(auth_type==S_enable)?"enableacl":"loginacl", (int)id->NAS_addr)) {
		res = store_result();
		if ((num_row=num_rows(res)) == 0) {
			if (taclog->debug)
				taclog->report("db::get_profile - NAS(%s) is not defined.", id->NAS_name->c_str());
			free_result(res);
			return (cont);
		}
		for (i=0;row=fetch_row(res);i++) {
			acl[i] = atoi(row[0]);
			if (i==0) *vendor=atoi(row[2]);
			if (!hostgroup.empty()&&hostgroup.compare(row[1]?row[1]:"")) {
				if (!query(HOSTGROUPSQL, (auth_type==S_enable)?"enableacl":"loginacl", hostgroup.c_str())) {
					res1 = store_result();
					if (num_rows(res1) > 0) {
						row1=fetch_row(res1);
						i++; num_row++;
						acl[i] = atoi(row1[0]);
					}
					free_result(res1);
				}
			}
			hostgroup.assign(row[1]?row[1]:"");
		}
		free_result(res);
		if (!hostgroup.empty()) {
			if (!query(HOSTGROUPSQL, (auth_type==S_enable)?"enableacl":"loginacl", hostgroup.c_str())) {
				res = store_result();
				if (num_rows(res) > 0) {
					row=fetch_row(res);
					acl[num_row] = atoi(row[0]);
					num_row++;
				}
				free_result(res);
			}
		}
		for (i=0;cont && (i<num_row);i++) {
			if (acl[i]) {
				if (!query(NASACLSQL, acl[i], id->username->c_str(), member.c_str())) {
					res = store_result();
					if (num_rows(res) > 0) {
						while (cont && (row=fetch_row(res))) {
							if (strcmp(row[0],"58")==0) {
								if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
									taclog->report("db::get_profile - user(%s) is not allowed to access the nas(%s).", id->username->c_str(), id->NAS_name->c_str());
								free_result(res);
								return (cont);
							} else if (strcmp(row[0],"57")==0)  {
								cont = 0;
								*profile = atol(row[1]);
							}
						}
					}
					free_result(res);
				} else {
					if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
						taclog->report("db::get_profile - problems with NASACLSQL query");
				}
			}
		}
		if (cont) {
			if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG))
				taclog->report("db::get_profile - user(%s) do not have permission to access this nas(%s)", id->username->c_str(), id->NAS_name->c_str());
		}
	} else {
	   if (taclog->debug & (DEBUG_PASSWD_FLAG|DEBUG_AUTHEN_FLAG|DEBUG_AUTHOR_FLAG))
		   taclog->report("db::get_profile - problems with HOSTACL query");
	}
	return (cont);
}

