/*
 *   Copyright (C) 2002-2006
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
Author:		Andrew Young (baram01@hotmail.com)
Last Updated:	October 21, 2007
 */

#include <iostream>           // For cout, cerr
#include <fstream>
#include <cstdlib>            // For atoi()  
#include <pthread.h>          // For POSIX threads  
#include <netinet/in.h>
#include <unistd.h>

#include "md5.h"
#include "tac_plus.h"
#include "tacacs.h"

#ifdef SYS_LINUX
#include <getopt.h>
#endif
#ifdef SYS_SOLARIS
#include <sys/socket.h>
#endif

Config *taccfg;
Log *taclog;

TACACSSocket::TACACSSocket(int newConnSD) : TCPSocket(newConnSD) {
	seq_no = 0;
}

u_char *TACACSSocket::read_packet()
{
	HDR hdr;
	u_char *pkt, *data;
	int len, dlen;
	
	if (taclog->debug)
		taclog->report("TACACSSocket::read_packet - Waiting for packet from %s on socket %u\n", getForeignAddress().c_str(), getLocalPort());

	if ((len = recv(&hdr, TAC_PLUS_HDR_SIZE, MSG_PEEK)) != TAC_PLUS_HDR_SIZE) {
		if (taclog->debug)
			taclog->report("TACACSSocket::read_packet - TACACAS Header is not correct\n");
		return (NULL);
	}

	if ((hdr.version & TAC_PLUS_MAJOR_VER_MASK) != TAC_PLUS_MAJOR_VER) {
		if (taclog->debug)
			taclog->report("TACACSSocket::read_packet - TACACS version is incorrect\n"); 
		return (NULL);
	}

	if ((hdr.type == TAC_PLUS_AUTHOR) || (hdr.type == TAC_PLUS_ACCT)) {
		if (hdr.version != TAC_PLUS_VER_0) {
			if (taclog->debug)
				taclog->report("TACACSSocket::read_packet - TACACS packet version is illegal\n");
			return (NULL);
		}
	}

	dlen = ntohl(hdr.datalength);
	len = TAC_PLUS_HDR_SIZE + dlen;
	if ((dlen & ~0xffffUL) || len > 0x10000) {
		if (taclog->debug)
			taclog->report("TACACSSocket::read_packet - Illegal date size: %u\n", ntohl(hdr.datalength));
		return (NULL);
	}

	/** reserve memory **/
	pkt = new u_char[len];

	/** data start here **/
	data = pkt + TAC_PLUS_HDR_SIZE;

	if (recv(pkt, len, 0) != len) {
		delete pkt;
		if (taclog->debug)
			taclog->report("TACACSSocket::read_packet - Bad socket read from: %s\n", getForeignAddress().c_str());
		return (NULL);
	}

	seq_no++;
	last_exch = time(NULL);

	if (seq_no != hdr.seq_no) {
		delete pkt;
		if (taclog->debug)
			taclog->report("TACACSSocket::read_packet - Illegal session seq # %u != packet seq # %u\n", seq_no, hdr.seq_no);
		return (NULL);
	}

	/** decrypt the data portion of packet **/
	if (md5_xor((HDR *)pkt, data, (char*)hkey.c_str())) {
		delete pkt;
		if (taclog->debug)
			taclog->report("TACACSSocket::read_packet - Cannot decrypt data\n");
		return (NULL);
	}

	if (taclog->debug)
		taclog->report("TACACSSocket::read_packet - Packet size=%u\n", len);

	if (taclog->debug & DEBUG_PACKET_FLAG) {
		HDR *hdr1 = (HDR *) pkt;
		if (taclog->debug & DEBUG_PACKET_FLAG) 
			taclog->report("TACACSSocket::read_packet - Dumping header\n>> PACKET: key=%s\n>> version %d (0x%x), type %d, seq no %d, encryption %d\nTACACSSocket::read_packet - End header\n", hkey.c_str(), hdr1->version, hdr1->version, hdr1->type, hdr1->seq_no, hdr1->encryption);
	}

	version = hdr.version;

	return (pkt);
}

int TACACSSocket::write_packet(u_char *pak)
{
	HDR *hdr = (HDR *) pak;
	u_char *data;
	int len;

	if (taclog->debug)
		taclog->report("TACACSSocket::write_packet - Sending packet to %s on socket %u\n", getForeignAddress().c_str(), getLocalPort());

	if (taclog->debug & DEBUG_PACKET_FLAG) {
		taclog->report("TACACSSocket::write_packet - Dumping header\n>> PACKET: key=%s\n>> version %d (0x%x), type %d, seq no %d, encryption %d\n", hkey.c_str(), hdr->version, hdr->version, hdr->type, hdr->seq_no, hdr->encryption);
		taclog->report("TACACSSocket::write_packet - End header\n");
	}

	len = TAC_PLUS_HDR_SIZE + ntohl(hdr->datalength);

	data = pak + TAC_PLUS_HDR_SIZE;

	/* encrypt the data portion */
	if (md5_xor((HDR *)pak, data, (char*)hkey.c_str())) {
		if (taclog->debug)
			taclog->report("TACACSSocket::write_packet - Cannot encrypt data\n");
		return (1);
	}

	if (send(pak, len, 0) != len) {
		if (taclog->debug)
			taclog->report("TACACSSocket::write_packet - Cannot send packet\n");
		return (1);
	}

	last_exch = time(NULL);
	return (0);
}

int TACACSSocket::send_acct_reply(u_char status, string *msg, string *data)
{
	u_char *pak, *p;
	HDR *hdr;
	int len, msg_len=0, data_len=0;
	struct acct_reply *reply;

	msg_len = (msg->empty())?0:msg->size();
	data_len = (data->empty())?0:data->size();

	len = TAC_PLUS_HDR_SIZE + TAC_ACCT_REPLY_FIXED_FIELDS_SIZE+msg_len+data_len;
	pak = new u_char[len];
	reply = (struct acct_reply *) (pak + TAC_PLUS_HDR_SIZE);
	hdr = (HDR *) pak;

	bzero(pak, len);

	hdr->version = TAC_PLUS_VER_0;
	hdr->type = TAC_PLUS_ACCT;
	hdr->seq_no = ++seq_no;
	hdr->encryption = TAC_PLUS_CLEAR;
	hdr->session_id = htonl(session_id);
	hdr->datalength = htonl(len - TAC_PLUS_HDR_SIZE);

	reply->status = status;
	reply->msg_len  = msg_len;
	reply->data_len = data_len;

	p = pak + TAC_PLUS_HDR_SIZE + TAC_ACCT_REPLY_FIXED_FIELDS_SIZE;
	if (msg_len) bcopy(msg->c_str(), p, msg_len);
	p += msg_len;

	if (msg_len) bcopy(data->c_str(), p, data_len);

	if (taclog->debug & DEBUG_PACKET_FLAG) {
		taclog->report("TACACSSocket::send_acct_reply: Writing ACCT size=%d\n", len);
//		dump_tacacs_pak(pak);
	}

	reply->msg_len = ntohs(reply->msg_len);
	reply->data_len = ntohs(reply->data_len);

	write_packet(pak);

	delete pak;
}

int TACACSSocket::send_author_reply(int status, string *msg, string *data, map<string,string>& args)
{
	u_char *pak, *p;
	HDR *hdr;
	struct author_reply *reply;
	int len;
	int i;

	//Calculate final packet size
	len = TAC_PLUS_HDR_SIZE + TAC_AUTHOR_REPLY_FIXED_FIELDS_SIZE + msg->size() +data->size();
/*
	for (i=0; i < arg_cnt; i++) {
		len += args[i]->size() + 1;
	}
*/
	map<string,string>::const_iterator iter;
	for (iter=args.begin(); iter!=args.end(); ++iter) {
		len += iter->first.size()+iter->second.size() + 1;
	}


	pak = new u_char[len];
	bzero(pak, len);
	hdr = (HDR *) pak;

	reply = (struct author_reply *) (pak + TAC_PLUS_HDR_SIZE);

	hdr->version = TAC_PLUS_VER_0;
	hdr->type = TAC_PLUS_AUTHOR;
	hdr->seq_no = ++seq_no;
	hdr->encryption = TAC_PLUS_CLEAR;
	hdr->session_id = htonl(session_id);
	hdr->datalength = htonl(len - TAC_PLUS_HDR_SIZE);

	reply->status   = status;
	reply->msg_len  = msg->size();
	reply->data_len = data->size();
	reply->arg_cnt  = args.size();

	p = pak + TAC_PLUS_HDR_SIZE + TAC_AUTHOR_REPLY_FIXED_FIELDS_SIZE;

	/* place arg sizes into packet  */
/*
	for (i=0; i < arg_cnt; i++) {
	    *p++ = args[i]->size();
	}
*/
	for (iter=args.begin(); iter!=args.end(); ++iter) {
		*p++ = iter->first.size()+iter->second.size();
	}

	if (msg->size()) bcopy(msg->c_str(), p, msg->size());
	p += msg->size();
	if (data->size()) bcopy(data->c_str(), p, data->size());
	p += data->size();

	/* copy arg bodies into packet */
/*
	for (i=0; i < arg_cnt; i++) {
	    int arglen = args[i]->size();

	    bcopy(args[i]->c_str(), p, arglen);
	    p += arglen;
	}
*/
	for (iter=args.begin(); iter!=args.end(); ++iter) {
		bcopy(iter->first.c_str(), p, iter->first.size());
		p += iter->first.size();
		bcopy(iter->second.c_str(), p, iter->second.size());
		p += iter->second.size();
	}

	if (taclog->debug & DEBUG_PACKET_FLAG) {
		char *pmsg;
		switch (status) {
		case AUTHOR_STATUS_PASS_ADD:
			pmsg = "AUTHOR/PASS_ADD";
			break;
		case AUTHOR_STATUS_FAIL:
			pmsg = "AUTHOR/FAIL";
			break;
		case AUTHOR_STATUS_PASS_REPL:
			pmsg = "AUTHOR/PASS_REPL";
			break;
		case AUTHOR_STATUS_ERROR:
			pmsg = "AUTHOR/ERROR";
			break;
		default:
			pmsg = "AUTHOR/UNKNOWN";
			break;
		}
		taclog->report("TACACSSocket::send_author_reply -Writing %s size=%d datasize=%d\n", pmsg, len, (len-TAC_PLUS_HDR_SIZE));
//		dump_tacacs_pak(pak);
	}

	reply->msg_len  = htons(reply->msg_len);
	reply->data_len = htons(reply->data_len);

	write_packet(pak);

	delete pak;
}

int TACACSSocket::send_authen_reply(int status, const char *msg, const char *data, int flags)
{
	u_char *pak, *p;
	HDR *hdr;
	struct authen_reply *reply;
	int msg_len, data_len, len;

	if (taclog->debug)
		taclog->report("TACACSSocket::send_authen_reply - message(%s) data(%s)\n", msg, data);

	data_len = (data)?strlen(data):0;
	msg_len = (msg)?strlen(msg):0;

	len = TAC_PLUS_HDR_SIZE + TAC_AUTHEN_REPLY_FIXED_FIELDS_SIZE + msg_len + data_len;

	pak = new u_char[len];
	bzero(pak, len);
	hdr = (HDR *) pak;
	reply = (struct authen_reply *) (pak + TAC_PLUS_HDR_SIZE);

	hdr->version = version;
	hdr->type = TAC_PLUS_AUTHEN;
	hdr->seq_no = ++seq_no;
	hdr->encryption = TAC_PLUS_CLEAR;
	hdr->session_id = htonl(session_id);
	hdr->datalength = htonl(TAC_AUTHEN_REPLY_FIXED_FIELDS_SIZE + msg_len + data_len);

	reply->status = status;
	reply->msg_len = msg_len;
	reply->data_len = data_len;
	reply->flags = flags;

	p = pak + TAC_PLUS_HDR_SIZE + TAC_AUTHEN_REPLY_FIXED_FIELDS_SIZE;

	if (msg_len) bcopy(msg, p, msg_len);
	p += msg_len;
	if (data_len) bcopy(data, p, data_len);

	if (taclog->debug & DEBUG_PACKET_FLAG) {
		string pmsg;
	        switch (status) {
	        case TAC_PLUS_AUTHEN_STATUS_PASS:
	            pmsg = "AUTHEN/SUCCEED";
	            break;
	        case TAC_PLUS_AUTHEN_STATUS_FAIL:
	            pmsg = "AUTHEN/FAIL";
	            break;
	        case TAC_PLUS_AUTHEN_STATUS_GETDATA:
	            pmsg = "AUTHEN/GETDATA";
	            break;
	        case TAC_PLUS_AUTHEN_STATUS_GETUSER:
	            pmsg = "AUTHEN/GETUSER";
	            break;
	        case TAC_PLUS_AUTHEN_STATUS_GETPASS:
	            pmsg = "AUTHEN/GETPASS";
	            break;
	        case TAC_PLUS_AUTHEN_STATUS_ERROR:
	            pmsg = "AUTHEN/ERROR";
	            break;
	        default:
	            pmsg = "AUTHEN/UNKNOWN";
	            break;
		}
		taclog->report("TACACSSocket::send_authen_reply: Writing %s size=%d\n", pmsg.c_str(), len);
//		dump_tacacs_pak(pak);
	}

	reply->msg_len = htons(reply->msg_len);
	reply->data_len = htons(reply->data_len);

	write_packet(pak);

	delete pak;
}

void TACACSSocket::send_authen_error(string msg)
{
	string buf;

	buf = getForeignAddress();
	buf+=" "+getForeignPort();
	buf+=": "+msg;

	if (taclog->debug | DEBUG_AUTHEN_FLAG)
		taclog->report("TACACSSocket::send_authen_error -%s\n", buf.c_str());

	send_authen_reply(TAC_PLUS_AUTHEN_STATUS_ERROR, buf.c_str(), (const char *)NULL, 0);
}
 
int TACACSSocket::get_authen_cont(u_char **reply)
{
        HDR *hdr;
        struct authen_cont *cont;
        string msg;
	u_char *pak;

        if ((*reply=read_packet())==NULL) {
		if (taclog->debug & DEBUG_AUTHEN_FLAG)
			taclog->report("TACACSSocket::get_authen_cont - Null reply packet, expecting CONTINUE\n");
                return (1);
        }

        hdr = (HDR *)*reply;
        cont = (struct authen_cont *) (*reply + TAC_PLUS_HDR_SIZE);

        if ((hdr->type != TAC_PLUS_AUTHEN) || (hdr->seq_no <= 1)) {
                msg="Bad packet type="+hdr->type;
		msg+="/seq no="+hdr->seq_no;
		msg+=" when expecting authentication cont";

                if (taclog->debug & DEBUG_AUTHEN_FLAG)
			taclog->report("TACACSSocket::get_authen_cont -%s\n",msg.c_str());
			send_authen_error(msg);
                delete *reply;
                return (1);
        }

        cont->user_msg_len = ntohs(cont->user_msg_len);
        cont->user_data_len = ntohs(cont->user_data_len);
        if ((TAC_AUTHEN_CONT_FIXED_FIELDS_SIZE + cont->user_msg_len + cont->user_data_len) != ntohl(hdr->datalength)) {
                msg="Illegal sized authentication continue packet";
                if (taclog->debug & DEBUG_AUTHEN_FLAG)
                        taclog->report("TACACSSocket::get_authen_cont -%s: %s\n", getForeignAddress().c_str(), msg.c_str());

                send_authen_error(msg);
                delete *reply;
                return (1);
        }

 //       if (debug & DEBUG_PACKET_FLAG)
//                dump_nas_pak(pak);

	if (taclog->debug)
        	taclog->report("TACACSSocket::get_authen_cont -%s: %d -- %s\n", getForeignAddress().c_str(), cont->flags);

        return (0);
}

TACACSServerSocket::TACACSServerSocket(unsigned short localPort, int queueLen) throw(SocketException) : TCPServerSocket(localPort, queueLen) {
}

TACACSSocket *TACACSServerSocket::accept() throw (SocketException)
{
	int newConnSD;
	if ((newConnSD = ::accept(sockDesc, NULL, 0)) < 0) {
		throw SocketException("TACACSSocket::accept -failed", true);
	}

	return new TACACSSocket(newConnSD);
}


void *ThreadMain(void *arg);               // Main program of a thread  

int main(int argc, char *argv[]) {
	unsigned short tacacsServPort;

	extern char *optarg;
	int c;

#ifdef TAC_PLUS_PORT
	tacacsServPort = TAC_PLUS_PORT;
#endif
	taccfg = new Config();
	taclog = new Log();

	while ((c = getopt(argc, argv, "d:C:p:Pvl:D:")) != EOF) {
		switch (c) {
		case 'C':
			taccfg->setconfigfile(optarg);
			break;
		case 'l':
			taclog->setlogfile(optarg);
			break;
		case 'd':
			taclog->debug = atoi(optarg);
			break;
		case 'p':
			tacacsServPort = atoi(optarg);
			break;
		case 'D':
			taccfg->dburl = new string(optarg);
			break;
		case 'v':
			cout << "tac_plus version " << VERSION_TAG << endl;
#ifdef DEBUG
			cout << "DEBUG" << endl;
#endif
#if defined(SYS_FREEBSD)
			cout << "FREEBSD" << endl;
#elif defined(SYS_LINUX)
			cout << "LINUX" << endl;
#elif defined(SYS_SOLARIS)
			cout << "SOLARIS" << endl;
#else
			cout << "UNKNOWN" << endl;
#endif
			return (1);
			break;
		case 'P':
			taccfg->parse();
			return (1);
			break;
		default:
			cerr << "Usage: " << argv[0] << " [-C <configuration file>] [-l <log file>]" << endl;
			cerr << "\t [-p <port>] [-d <debug level>]" << endl;
			cerr << "\t [-D <dburl>] [-v] [-P]" << endl;
			return (1);
		}
	}

	taclog->init();
	taccfg->loadcfg();

	//Write process id to pidfile
	fstream pid("/tmp/tac_plus.pid", fstream::out | fstream::app);
	pid << getpid();
	pid.close();

	try {
	// Socket descriptor for server
		TACACSServerSocket servSock(tacacsServPort);
  
		for (;;) {      // Run forever  
			taclog->report("Waiting for connection\n");
			
 			// Create client thread  
			pthread_t threadID;
		        if (pthread_create(&threadID, NULL, ThreadMain, (void *) servSock.accept()) != 0) { 
				taclog->report("main: Unable to create thread\n");
			}
		}
	} catch (SocketException &e) {
		taclog->report("%s\n",e.what());
	}

	taclog->report("Closing all connections\n");

	return 0;
}


void *ThreadMain(void *sock) {
	TACACSSocket *s = (TACACSSocket*)sock;
	tacacs *svr = new tacacs();
	pthread_t ThreadId_ = pthread_self();

	taclog->report("ThreadMain - Handling client %s with thread %u\n", s->getForeignAddress().c_str(), ThreadId_);
	pthread_detach(ThreadId_);
	svr->Execute(s);
	taclog->report("ThreadMain - thread %u ending\n", s->getForeignAddress().c_str());
	delete svr;
	delete s;
	return NULL;
}


