/*
 *   Copyright (C) 2002-2006
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _TAC_PLUS_H
#define _TAC_PLUS_H

#include <map>
#include <string>

using namespace std;

#include "config.h"
#include "PracticalSocket.h" 
#include "Thread.h"
#include "Log.h"

#ifdef SYS_SOLARIS
#include <strings.h>
#endif

#define VERSION_TAG VERSION " (Extended Tacplus)"
#define PROMPT_INSERT "\nUser Access Verification (" VERSION ")\nUsername: "

#define MD5_LEN			16

#ifdef  MSCHAP
#define MSCHAP_DIGEST_LEN	49
#endif

#define TAC_PLUS_PORT		49
#define TAC_PLUS_READ_TIMEOUT	180
#define TAC_PLUS_WRITE_TIMEOUT	180
#define NAS_PORT_MAX_LEN	255
#define TAC_PLUS_MAX_ITERATIONS	50
#define MAX_INPUT_LINE_LEN	255

#define TACPLUS_PIDFILE		"/var/run/tac_plus.pid"
#define LOGFILE_DEFAULT		"/var/log/tac_plus.log"

#ifndef NULL
#define NULL			((void*) 0)
#endif

#undef  MAX
#define MAX(a, b)		(((a) > (b)) ? (a) : (b))

#undef  MIN
#define MIN(a, b)		(((a) < (b)) ? (a) : (b))

#define DOLLARSIGN		'$'

#define DEFAULT_USERNAME	"DEFAULT"

/* Debugging flags */
#define DEBUG_PARSE_FLAG     2
#define DEBUG_FORK_FLAG      4
#define DEBUG_AUTHOR_FLAG    8
#define DEBUG_AUTHEN_FLAG    16
#define DEBUG_PASSWD_FLAG    32
#define DEBUG_ACCT_FLAG      64
#define DEBUG_CONFIG_FLAG    128
#define DEBUG_PACKET_FLAG    256
#define DEBUG_HEX_FLAG       512
#define DEBUG_MD5_HASH_FLAG  1024
#define DEBUG_XOR_FLAG       2048
#define DEBUG_CLEAN_FLAG     4096
#define DEBUG_SUBST_FLAG     8192
#define DEBUG_PROXY_FLAG     16384
#define DEBUG_MAXSESS_FLAG   32768
#define DEBUG_LOCK_FLAG      65536

/* Authentication types */
#define S_license	  2
#define S_login           8
#define S_arap            13
#define S_chap            14
#define S_pap             16
#define S_svc             17
#define S_ppp             29
#define S_skey            31
#define S_mschap          42
#define S_db              44
#define S_ldap            47
#define S_enable          50

#define PW_OK             0
#define PW_EXPIRED        1
#define PW_EXPIRING       2
#define WARNING_PERIOD    7
#define MAX_PASSWD_LEN    256

/*
 * XTACACSP protocol definitions
 */

#define AUTHEN_NAME_SIZE	128

#define CHOOSE_FAILED	-1   /*failed to choose an authentication function*/
#define CHOOSE_OK	0    /*successfully chose an authentication function*/
#define CHOOSE_GETUSER	1    /*need a username before choosing*/
#define CHOOSE_BADTYPE	2    /*Invalid preferred authen function specified*/

/*
   authen_func
*/
#define DEFAULT_V0_FN	1
#define DEFAULT_FN	2
#define ENABLE_FN	3
#define SENDAUTH_FN	4
#define SENDPASS_FN	5

/*
 * This structure describes an authentication method.
 *   authen_name     contains the name of the authentication method.
 *   authen_func     is a pointer to the authentication function.
 *   authen_method   numeric value of authentication method
 */

struct authen_type {
	int authen_func;
	int authen_type;
};

/*
 * This structure describes a principal that is to be authenticated.
 *   username        is the principals name (ASCII, null terminated)
 *   NAS_name        is the name of the NAS where the user is
 *   NAS_port        is the port on the NAS where the user is
 *   NAS_addr            is the ip address of the NAS where the user is
 *   NAC_address     is the remote user location.  This may be
 *                   a remote IP address or a caller-ID or ...
 *   priv_lvl        user's requested privilege level.
 */

struct identity {
    string              *username;
    string              *NAS_name;
    string              *NAS_port;
    uint32_t            NAS_addr;
    string              *NAC_address;
    int                 priv_lvl;
};

/* internal state variables */
#define STATE_AUTHEN_START   0  /* no requests issued */
#define STATE_AUTHEN_GETUSER 1  /* username has been requested */
#define STATE_AUTHEN_GETPASS 2  /* password has been requested */

struct private_data {
	string *password;
	int state;
};

/*
 * The authen_data structure is the data structure for passing
 * information to and from the authentication function
 * (authen_type.authen_func).
 */

struct authen_data {
    struct identity *NAS_id;    /* user identity */
//    string *server_msg;           /* null-terminated output msg */

 //   string *server_data;

    char *client_msg;           /* null-terminated input msg a user typed */

    int client_dlen;            /* input data length */
    char *client_data;          /* input data */

    int state;                  /* Password state */
    int action;                 /* what's to be done */
    int service;                /* calling service */
    int status;                 /* Authen status */
    int type;                   /* Authen type */
    u_char flags;               /* input & output flags fields */
};

typedef struct authen_data AUTHEN_DATA;

/*
 * This structure is the data structure for passing information to
 * and from the authorization function (do_author()).
 */
#define AUTHEN_METH_NONE             0x01
#define AUTHEN_METH_KRB5             0x02
#define AUTHEN_METH_LINE             0x03
#define AUTHEN_METH_ENABLE           0x04
#define AUTHEN_METH_LOCAL            0x05
#define AUTHEN_METH_TACACSPLUS       0x06
#define AUTHEN_METH_RCMD             0x20

#define AUTHOR_STATUS_PASS_ADD       0x01
#define AUTHOR_STATUS_PASS_REPL      0x02
#define AUTHOR_STATUS_FAIL           0x10
#define AUTHOR_STATUS_ERROR          0x11

/* Authorization service type */
#define N_arg           50
#define N_optarg        51
#define N_svc_exec      52
#define N_svc_slip      53
#define N_svc_ppp       54
#define N_svc_arap      55
#define N_svc_cmd       56
#define N_permit        57
#define N_deny          58
#define N_svc           59

struct author_data {
    struct identity *id;        /* user id */
    int authen_method;          /* authentication method */
    int authen_type;            /* authentication type see authen_type */
    int service;                /* calling service */
    string *msg;                  /* optional NULL-terminated return message */
    string *admin_msg;            /* optional NULL-terminated admin message */
    int status;                 /* return status */
//    int num_in_args;            /* input arg count */
//    char **input_args;          /* input arguments */
//    int num_out_args;           /* output arg cnt */
//    char **output_args;         /* output arguments */
};

/* An API accounting record structure */

#define ACCT_TYPE_START     	 1
#define ACCT_TYPE_STOP      	 2
#define ACCT_TYPE_UPDATE    	 3

struct acct_rec {
    int acct_type;              /* start, stop, update */
    struct identity *identity;
    int authen_method;
    int authen_type;
    int authen_service;
    string msg;                  /* output field */
    string admin_msg;            /* output field */
//    int num_args;
//    char **args;
};

/*
 TACACS+ Header
*/
#define TAC_PLUS_MAJOR_VER_MASK		0xf0
#define TAC_PLUS_MAJOR_VER		0xc0
#define TAC_PLUS_MINOR_VER_0		0x0
#define TAC_PLUS_VER_0			(TAC_PLUS_MAJOR_VER | TAC_PLUS_MINOR_VER_0)
#define TAC_PLUS_MINOR_VER_1		0x01
#define TAC_PLUS_VER_1			(TAC_PLUS_MAJOR_VER | TAC_PLUS_MINOR_VER_1)

#define TAC_PLUS_AUTHEN			1
#define TAC_PLUS_AUTHOR			2
#define TAC_PLUS_ACCT			3

#define TAC_PLUS_ENCRYPTED		0x0	/* packet encrypted */
#define TAC_PLUS_CLEAR			0x1	/* packet not encrypted */

#define TAC_PLUS_HDR_SIZE		12
struct tac_plus_pak_hdr {
	u_char version;
	u_char type;
	u_char seq_no;		/* packet sequence number */
	u_char encryption;	/* packet is encrypted or cleartext */
	int session_id;
	int datalength;
};

typedef struct tac_plus_pak_hdr		 HDR;

#define HASH_TAB_SIZE			127  /* hash table size */

/* Authentication packet sent from NAS */
#define TAC_PLUS_AUTHEN_LOGIN   	 0x1
#define TAC_PLUS_AUTHEN_CHPASS  	 0x2
#define TAC_PLUS_AUTHEN_SENDPASS	 0x3 /* deprecated */
#define TAC_PLUS_AUTHEN_SENDAUTH	 0x4

#define TAC_PLUS_PRIV_LVL_MIN		 0x0
#define TAC_PLUS_PRIV_LVL_MAX		 0xf

#define TAC_PLUS_AUTHEN_TYPE_ASCII 	 1
#define TAC_PLUS_AUTHEN_TYPE_PAP   	 2
#define TAC_PLUS_AUTHEN_TYPE_CHAP  	 3
#define TAC_PLUS_AUTHEN_TYPE_ARAP  	 4
#define TAC_PLUS_AUTHEN_TYPE_MSCHAP	 5

#define TAC_PLUS_AUTHEN_SVC_LOGIN 	 1
#define TAC_PLUS_AUTHEN_SVC_ENABLE	 2
#define TAC_PLUS_AUTHEN_SVC_PPP   	 3
#define TAC_PLUS_AUTHEN_SVC_ARAP  	 4
#define TAC_PLUS_AUTHEN_SVC_PT    	 5
#define TAC_PLUS_AUTHEN_SVC_RCMD  	 6
#define TAC_PLUS_AUTHEN_SVC_X25   	 7
#define TAC_PLUS_AUTHEN_SVC_NASI  	 8

#define TAC_AUTHEN_START_FIXED_FIELDS_SIZE 8
struct authen_start {
	u_char action;
	u_char priv_lvl;
	u_char authen_type;
	u_char service;
	u_char user_len;
	u_char port_len;
	u_char rem_addr_len;
	u_char data_len;
};

/* Authentication continue packet from the NAS */

#define TAC_PLUS_CONTINUE_FLAG_ABORT	0x1

#define TAC_AUTHEN_CONT_FIXED_FIELDS_SIZE	5

struct authen_cont {
	u_short user_msg_len;
	u_short user_data_len;
	u_char flags;
};

/* Authentication reply packet sent to the NAS */
#define TAC_PLUS_AUTHEN_STATUS_PASS     1
#define TAC_PLUS_AUTHEN_STATUS_FAIL     2
#define TAC_PLUS_AUTHEN_STATUS_GETDATA  3
#define TAC_PLUS_AUTHEN_STATUS_GETUSER  4
#define TAC_PLUS_AUTHEN_STATUS_GETPASS  5
#define TAC_PLUS_AUTHEN_STATUS_RESTART  6
#define TAC_PLUS_AUTHEN_STATUS_ERROR    7
#define TAC_PLUS_AUTHEN_STATUS_FOLLOW   0x21

#define TAC_PLUS_AUTHEN_FLAG_NOECHO     0x1

#define TAC_AUTHEN_REPLY_FIXED_FIELDS_SIZE 6
struct authen_reply {
	u_char status;
	u_char flags;
	u_short msg_len;
	u_short data_len;
};

/* Authorization packet sent from NAS */
#define TAC_AUTHOR_REQ_FIXED_FIELDS_SIZE 8
struct author {
    u_char authen_method;
    u_char priv_lvl;
    u_char authen_type;
    u_char service;

    u_char user_len; u_char port_len;
    u_char rem_addr_len;
    u_char arg_cnt;             /* the number of args */
};

/* Authorization packet sent to the NAS */
#define TAC_AUTHOR_REPLY_FIXED_FIELDS_SIZE 6
struct author_reply {
    u_char status;
    u_char arg_cnt;
    u_short msg_len;
    u_short data_len;
};

/* Accounting */
#define TAC_PLUS_ACCT_FLAG_MORE     0x1
#define TAC_PLUS_ACCT_FLAG_START    0x2
#define TAC_PLUS_ACCT_FLAG_STOP     0x4
#define TAC_PLUS_ACCT_FLAG_WATCHDOG 0x8

#define TAC_ACCT_REQ_FIXED_FIELDS_SIZE 9

struct acct {
    u_char flags;
    u_char authen_method;
    u_char priv_lvl;
    u_char authen_type;
    u_char authen_service;
    u_char user_len;
    u_char port_len;
    u_char rem_addr_len;
    u_char arg_cnt; /* the number of cmd args */
};

/* Accounting reply */
#define TAC_PLUS_ACCT_STATUS_SUCCESS 0x1
#define TAC_PLUS_ACCT_STATUS_ERROR   0x2
#define TAC_PLUS_ACCT_STATUS_FOLLOW  0x21

#define TAC_ACCT_REPLY_FIXED_FIELDS_SIZE 5
struct acct_reply {
    u_short msg_len;
    u_short data_len;
    u_char status;
};

class TACACSSocket : public TCPSocket
{
	int seq_no;
	time_t last_exch;

public:
	string hkey;
	uint32_t session_id;
	u_char version;

	TACACSSocket(int newConnSD);
	u_char *read_packet();
	int write_packet(u_char *pak);

	void send_authen_error(string msg);
	int send_authen_reply(int status, const char *msg, const char *data, int flags);
	int send_author_reply(int status, string *msg, string *data, map<string,string>& args);
	int send_acct_reply(u_char status, string *msg, string *data);

//	u_char *get_authen_cont();
	int get_authen_cont(u_char **reply);
};

class TACACSServerSocket : public TCPServerSocket
{
public:
	TACACSServerSocket(unsigned short localPort, int queueLen = 5) throw(SocketException);
	TACACSSocket *accept() throw(SocketException);
};

class tauthen;
class tauthor;

//extern char *make_str(u_char *p, int len);
template<class K, class V> void getargs(u_char *p, u_char *argsize, int cnt, map<K,V>& args);
extern char *find_attr_value(char *attr, char **args, int cnt);
extern int SD_Authenticate(string user, string user_passwd);

#include "file.h"
#include "db.h"
#include "Config.h"

extern Config *taccfg;
extern Log *taclog;
#endif
