/*
 *   Copyright (C) 2002-2006
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cstdlib>
#include <netinet/in.h>

#include "tac_plus.h"
#include "tacacs.h"

tacacs::tacacs() {
	dbh = new db(taccfg->dburl);
	fileh = new file(taccfg->acctfile);
}

tacacs::~tacacs() {
	delete dbh;
	delete fileh;
}

void tacacs::Execute (TACACSSocket *arg) 
{
	sock = arg;

        //Do lookup in configuration for hkey
	switch(taccfg->cfgtype) {
	case USEFILE:
		sock->hkey.assign("utopia");
		break;
	case USEDB:
		if (dbh->get_host("hkey", sock->getForeignuAddress(), &sock->hkey) > 0) {
			taclog->report("tacacs::Execute - Cannot find host (%s)\n", sock->getForeignAddress().c_str());
			return;
		}
		break;
	default:
		taclog->report("tacacs::Execute - Cannot find type\n");
	}

	//Check if hkey is not NULL or blank
        if (sock->hkey.size()<=0) {
                return;
        }

        if ((pak = sock->read_packet())==NULL) {
                return;
        }

        hdr = (HDR *)pak;
        sock->session_id = ntohl(hdr->session_id);

        switch (hdr->type) {
        case TAC_PLUS_AUTHEN:
		authen();
		break;
	case TAC_PLUS_AUTHOR:
		author();
		break;
	case TAC_PLUS_ACCT:
		accounting();
		break;
	default:
		taclog->report("tacacs::Execute -Illegal type %d in received packet\n", hdr->type);
	}

	delete pak;
}

void tacacs::authen()
{
	struct identity identity;
	struct authen_start *start;
	tauthen *authen = new tauthen(this);
	u_char *p;

	if (taclog->debug & DEBUG_AUTHEN_FLAG)
		taclog->report("tacacs::authen -Start authentication\n");

	start = (struct authen_start *) (pak + TAC_PLUS_HDR_SIZE);

	if ((hdr->seq_no != 1) || (ntohl(hdr->datalength) != TAC_AUTHEN_START_FIXED_FIELDS_SIZE + start->user_len + start->port_len + start->rem_addr_len + start->data_len)) {
		sock->send_authen_error("Invalid AUTHEN/START packet (check keys)\n");
		return;
	}

	switch (start->action) {
	case TAC_PLUS_AUTHEN_LOGIN:
	case TAC_PLUS_AUTHEN_SENDAUTH:
	case TAC_PLUS_AUTHEN_SENDPASS:
		break;

	default:
		string msg;
		msg = "Invalid AUTHEN/START action=";
		msg += start->action + "\n";
		sock->send_authen_error(msg);
		return;
	} 

	// variable length data starts here
	p = pak + TAC_PLUS_HDR_SIZE + TAC_AUTHEN_START_FIXED_FIELDS_SIZE;

	// identity structure
        identity.username = new string(string((char*)p).substr(0,start->user_len));
        p += start->user_len;
        identity.NAS_name = new string(sock->getForeignAddress());
        identity.NAS_addr = sock->getForeignuAddress();
        identity.NAS_port = new string(string((char*)p).substr(0,start->port_len));
        p += start->port_len;
        identity.NAC_address = new string(string((char*)p).substr(0,start->rem_addr_len));
        p += start->rem_addr_len;
        identity.priv_lvl = start->priv_lvl;

	// authen_data structure
	authen->id		 = &identity;
	authen->action      = start->action;
	authen->service     = start->service;
	authen->atype       = start->authen_type;
	authen->client_dlen = start->data_len;
	authen->client_data = new char[start->data_len];
	bcopy(p, authen->client_data, start->data_len);

	// authen_type structure
	authen->type->authen_type = start->authen_type;

	if (!authen->choose()) {
		authen->authenticate();
	}

	//freeing up identity structure
        if (identity.username) delete identity.username;
        if (identity.NAS_name) delete identity.NAS_name;
        if (identity.NAS_port) delete identity.NAS_port;
        if (identity.NAC_address) delete identity.NAC_address;
	delete authen;

	if (taclog->debug & DEBUG_AUTHEN_FLAG)
		taclog->report("tacacs::authen -Ending Authentication\n");
}

void tacacs::author()
{
	struct identity identity;
	struct author *start;
	tauthor *author = new tauthor(this);
	u_char *p, *argsizep;
	int len;

	if (taclog->debug & DEBUG_AUTHOR_FLAG)
		taclog->report("tacacs::author -Start authorizing\n");

	start = (struct author *)(pak + TAC_PLUS_HDR_SIZE);

	// arg counts start here
	p = pak + TAC_PLUS_HDR_SIZE + TAC_AUTHOR_REQ_FIXED_FIELDS_SIZE;

	len = TAC_AUTHOR_REQ_FIXED_FIELDS_SIZE;
	len += start->user_len + start->port_len + start->rem_addr_len + start->arg_cnt;
	for (int i=0; i < (int)start->arg_cnt; i++) {
		len += p[i];
	}

	if (len != ntohl(hdr->datalength)) {
		string msg = "";
		string adm = "";
		sock->send_author_reply(AUTHOR_STATUS_ERROR, &msg, &adm, author->output_args);
		if (taclog->debug & DEBUG_AUTHOR_FLAG)
			taclog->report("tacacs::author -Check keys\n");
		return;
	}

	argsizep = p;
	p += start->arg_cnt;

        // identity structure
        identity.username = new string(string((char*)p).substr(0,start->user_len));
        p += start->user_len;
        identity.NAS_name = new string(sock->getForeignAddress());
        identity.NAS_addr = sock->getForeignuAddress();
        identity.NAS_port = new string(string((char*)p).substr(0,start->port_len));
        p += start->port_len;
        identity.NAC_address = new string(string((char*)p).substr(0,start->rem_addr_len));
        p += start->rem_addr_len;
        identity.priv_lvl = start->priv_lvl;

	author->id = &identity;
	author->authen_method = start->authen_method;
	author->authen_type = start->authen_type;
	author->service = start->service;
//	author->num_in_args = start->arg_cnt;
//	author->input_args = (string**) new string[start->arg_cnt];

	getargs(p, argsizep, start->arg_cnt, author->input_args);

/*
	for (int i=0; i < (int) start->arg_cnt; i++) {
		author->input_args[i] = new string(string((char*)p).substr(0,*argsizep));
//		author.input_args[i] = make_str(p, *argsizep);
		p += *argsizep++;
	}
*/
	author->authorize();

	if (taclog->debug & DEBUG_AUTHOR_FLAG) {
		taclog->report("tacacs::author - %s@%s:%s %s\n",
				identity.username->c_str(), identity.NAS_name->c_str(),
				identity.NAS_port->c_str(),
				(author->status == AUTHOR_STATUS_PASS_ADD ||
				 author->status == AUTHOR_STATUS_PASS_REPL) ? "accepted" : "rejected");
	}

        //freeing up identity structure
        delete identity.username;
        delete identity.NAS_name;
        delete identity.NAS_port;
        delete identity.NAC_address;
	delete author;
                                                                                
        if (taclog->debug & DEBUG_AUTHOR_FLAG)
                taclog->report("tacacs::author -Ending Authorization\n");
}

void tacacs::accounting()
{
	struct acct *acct_pak;
	struct acct_rec rec;
	struct identity identity;

	map<string,string> args;
	u_char *p, *argsizep;
	int len, status, skipmax=0, current=0;
	string *pstr;

	if (taclog->debug & DEBUG_ACCT_FLAG)
		taclog->report("tacacs::accounting -Start accounting\n");

	if (hdr->version != TAC_PLUS_VER_0) {
		string msg = "Illegal accounting packet version";
		string adm = "";
		sock->send_acct_reply(TAC_PLUS_ACCT_STATUS_ERROR, &msg, &adm);
		return;
	}

	acct_pak = (struct acct *) (pak + TAC_PLUS_HDR_SIZE);
	//Do sanity checking

	/* arg counts start here */
	p = pak + TAC_PLUS_HDR_SIZE + TAC_ACCT_REQ_FIXED_FIELDS_SIZE;

	/* Length checks */
	len = TAC_ACCT_REQ_FIXED_FIELDS_SIZE;
	len += acct_pak->user_len + acct_pak->port_len +
		acct_pak->rem_addr_len + acct_pak->arg_cnt;
	for (int i = 0; i < (int)acct_pak->arg_cnt; i++) {
		len += p[i];
	}

	if (len != ntohl(hdr->datalength)) {
		string msg = "Illegal hkey";
		string adm = "";
		sock->send_acct_reply(TAC_PLUS_ACCT_STATUS_ERROR, &msg, &adm);
		if (taclog->debug & DEBUG_ACCT_FLAG)
			taclog->report("tauthen::accounting - Finish accouting (check keys)\n");
		return;
	}

	if (acct_pak->flags & TAC_PLUS_ACCT_FLAG_WATCHDOG)
	            rec.acct_type = ACCT_TYPE_UPDATE;
	if (acct_pak->flags & TAC_PLUS_ACCT_FLAG_START)
	            rec.acct_type = ACCT_TYPE_START;
	if (acct_pak->flags & TAC_PLUS_ACCT_FLAG_STOP)
	            rec.acct_type = ACCT_TYPE_STOP;

	rec.authen_method  = acct_pak->authen_method;
	rec.authen_type    = acct_pak->authen_type;
	rec.authen_service = acct_pak->authen_service;

	/* start of variable length data is here */
	p = pak + TAC_PLUS_HDR_SIZE + TAC_ACCT_REQ_FIXED_FIELDS_SIZE;

	/* skip arg cnts */
	p += acct_pak->arg_cnt;

	/* zero out identity struct */
	bzero(&identity, sizeof(struct identity));

	pstr = new string((char*)p);
	identity.username=new string(pstr->substr(current,acct_pak->user_len));
	p += acct_pak->user_len;
	current += (int)acct_pak->user_len;

	identity.NAS_name=new string(sock->getForeignAddress());
	identity.NAS_addr = sock->getForeignuAddress();

	identity.NAS_port=new string(pstr->substr(current,acct_pak->port_len));
	p += acct_pak->port_len;
	current += acct_pak->port_len;

	identity.NAC_address=new string(pstr->substr(current,acct_pak->rem_addr_len));
	p += acct_pak->rem_addr_len;
	current += acct_pak->rem_addr_len;

	identity.priv_lvl = acct_pak->priv_lvl;

	rec.identity = &identity;
	delete pstr;
	/* Now process cmd args */
	argsizep = pak + TAC_PLUS_HDR_SIZE + TAC_ACCT_REQ_FIXED_FIELDS_SIZE;

	getargs(p, argsizep, acct_pak->arg_cnt, args);

	if (taclog->debug & DEBUG_ACCT_FLAG) {
		map<string,string>::const_iterator iter;
		for (iter=args.begin(); iter!=args.end(); ++iter) {
			taclog->report("args[%s]=%s\n",iter->first.c_str(), iter->second.c_str());
		}
	}

	switch (taccfg->cfgtype) {
	case USEFILE:
		fileh->acct(&rec, args);
		status = TAC_PLUS_ACCT_STATUS_SUCCESS;
		break;
	case USEDB:
		string querycmd(ACCTSQL2);

		querycmd += "uid='"+*rec.identity->username;
		querycmd += "', nas='"+*rec.identity->NAS_name;
		querycmd += "', client_ip='"+*rec.identity->NAC_address;
		querycmd += "', terminal='"+*rec.identity->NAS_port;

		switch(rec.acct_type) {
		case ACCT_TYPE_UPDATE:
			querycmd += "', type='update'";
			break;
		case ACCT_TYPE_START:
			querycmd += "', type='start'";
			break;
		case ACCT_TYPE_STOP:
			querycmd += "', type='stop'";
			break;
		default:
			querycmd += "', type='unknown'";
		}

		if (!args["service"].empty()) {
			querycmd += ", service='";
			querycmd += args["service"].substr(1);
			querycmd += "'";
		}
		if (!args["cmd"].empty()) {
			querycmd += ", cmd='";
			querycmd += args["cmd"].substr(1);
			querycmd += "'";
		}
		if (!args["elapsed_time"].empty()) {
			querycmd += ", elapsed_time";
			querycmd += args["elapsed_time"];
		}
		if (!args["bytes_in"].empty()) {
			querycmd += ", bytes_in";
			querycmd += args["bytes_in"];
		}
		if (!args["bytes_out"].empty()) {
			querycmd += ", bytes_out";
			querycmd += args["bytes_out"];
		}
		if (!args["priv-lvl"].empty()) {
			querycmd += ", priv_lvl";
			querycmd += args["priv-lvl"];
		}

		if (dbh->acct(&querycmd))
			status = TAC_PLUS_ACCT_STATUS_ERROR;
		else
			status = TAC_PLUS_ACCT_STATUS_SUCCESS;
		break;
	}
	sock->send_acct_reply(status, &rec.msg, &rec.admin_msg);

	if (!identity.username->empty()) delete identity.username;
	if (!identity.NAS_name->empty()) delete identity.NAS_name;
	if (!identity.NAS_port->empty()) delete identity.NAS_port;
	if (!identity.NAC_address->empty()) delete identity.NAC_address;

	if (taclog->debug & DEBUG_ACCT_FLAG)
		taclog->report("tauthen::accounting - Finish accouting\n");
}

template<class K, class V>
void getargs(u_char *p, u_char *argsize, int cnt, map<K,V>& args)
{
	int ret = 1,
	    current = 0;

	if (cnt>0) {
		string *pargs = new string((char*)p);
		for (int i=0; i<cnt; i++) {
			string *argp = new string(pargs->substr(current,(int)*argsize));
			current += (int)*argsize;
			p += *argsize++;

			if (!argp->empty()) {
				int index;

				if ((index=argp->find("="))<=0) {
					index = argp->find("*");
				}	
				if (args[argp->substr(0,index)].empty())
					args[argp->substr(0,index)] = string(argp->substr(index));
				else
					args[argp->substr(0,index)] += string(" ") + string(argp->substr(index+1));
			}
			delete argp;
		}
		delete pargs;
	}
}

